﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.IO;

namespace Understanding_Linq_2
{
    public class CsvFileReader<TData> : IEnumerable<TData>, IDisposable where TData : new()
    {
        public CsvFileReader(TextReader stream)
        {
            UnderlyingStream = stream;
            InitColumnMappings();
        }

        /// <summary>
        /// Analyze properties of TData and look for the CsvColumnAttribute
        /// </summary>
        private void InitColumnMappings()
        {
            m_columnMappings = new PropertyInfo[0];

            foreach (PropertyInfo propInfo in typeof(TData).GetProperties(BindingFlags.Instance | BindingFlags.Public))
            {
                CsvColumnAttribute[] columnInfo = (CsvColumnAttribute[])propInfo.GetCustomAttributes(typeof(CsvColumnAttribute), true);
                if ((columnInfo != null) && (columnInfo.Length == 1))
                {
                    int targetIndex = columnInfo[0].ColumnIndex;
                    if (m_columnMappings.Length <= targetIndex) Array.Resize(ref m_columnMappings, targetIndex + 1);
                    m_columnMappings[targetIndex] = propInfo;
                }
            }
        }

        /// <summary>
        /// Fill an empty TData from a line of a csv file.
        /// </summary>
        /// <param name="line"></param>
        /// <param name="data"></param>
        private void FillObject(string line, TData data)
        {
            //split columns
            string[] parts = line.Split(';');

            //load new object
            for (int idx = 0; idx < m_columnMappings.Length; idx++)
            {
                PropertyInfo targetProperty = m_columnMappings[idx];
                if (targetProperty != null)
                {
                    targetProperty.SetValue(data, Convert.ChangeType(parts[idx], targetProperty.PropertyType), null);
                }
            }
        }

        /// <summary>
        /// Array of the properties to load from file in the same order as the columns they are to be filled from.
        /// </summary>
        PropertyInfo[] m_columnMappings;
        public TextReader UnderlyingStream { get; private set; }

        #region IEnumerable<TData> Members

        public IEnumerator<TData> GetEnumerator()
        {
            //skip header
            UnderlyingStream.ReadLine();

            string line;
            while ((line = UnderlyingStream.ReadLine()) != null)
            {
                line = line.Trim();

                if (line.Length > 0)
                {
                    TData data = new TData();

                    FillObject(line, data);

                    yield return data;
                }
            }
        }

        #endregion

        #region IEnumerable Members

     
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            if (UnderlyingStream != null) UnderlyingStream.Dispose();
        }

        #endregion


        public Filter<TData> Where(Func<TData, bool> filter)
        {
            return new Filter<TData>(this, filter);
        }
    }
}
