﻿using System;
using System.Collections.Generic;
using System.Linq.Expressions;
using System.Reflection;

namespace Understanding_Linq_3a
{
    public class Filter<TData> : IEnumerable<TData> where TData : new()
    {
        public Filter(CsvFileReader<TData> source, Expression<Func<TData, bool>> filter)
        {
            m_filter = filter;
            m_source = source;
        }

        CsvFileReader<TData> m_source;
        Expression<Func<TData, bool>> m_filter;

        internal IEnumerable<TData> GetRemote(RemoteExecutor executor)
        {
            executor.Filter = ExtractFilter();
            return m_source.GetRemote(executor);
        }

        private string ExtractFilter()
        {
            int filterColumnIndex;
            string filterOperator;
            string filterValue;
            string filterType;

            //outer must always be lambda; get the body which must be binary expression
            BinaryExpression binaryExpr = (BinaryExpression)m_filter.Body;

            //extract operator
            switch (binaryExpr.NodeType)
            {
                case ExpressionType.Equal:
                    filterOperator = "==";
                    break;
                case ExpressionType.LessThan:
                    filterOperator = "<";
                    break;
                case ExpressionType.GreaterThan:
                    filterOperator = ">";
                    break;
                default:
                    throw new ArgumentException("Filter", "Unsupported filter expression");
            }

            Expression left = binaryExpr.Left;

            //now left one must be member access of parameter
            if (left.NodeType != ExpressionType.MemberAccess) throw new ArgumentException("Filter", "Unsupported filter expression");
            MemberExpression leftAsMember = left as MemberExpression;
            if (leftAsMember.Expression.NodeType != ExpressionType.Parameter) throw new ArgumentException("Filter", "Unsupported filter expression");
            //remember datatype
			if (!(leftAsMember.Member is PropertyInfo)) throw new ArgumentException("Filter", "Unsupported filter expression");
            filterType = (leftAsMember.Member as PropertyInfo).PropertyType.AssemblyQualifiedName;

            //get the column index 
            filterColumnIndex = (leftAsMember.Member.GetCustomAttributes(typeof(CsvColumnAttribute), true) as CsvColumnAttribute[])[0].ColumnIndex;

            //right must evaluate to constant; use dirty trick with a parameterless lambda
            LambdaExpression rightLambda = Expression.Lambda(binaryExpr.Right);
            filterValue = Convert.ChangeType(rightLambda.Compile().DynamicInvoke(), Type.GetType(filterType)).ToString();


            return string.Format("{0};{1};{2};{3}", filterColumnIndex, filterOperator, filterValue, filterType);
        }
        
        #region IEnumerable<TData> Members

        public IEnumerator<TData> GetEnumerator()
        {
            RemoteExecutor executor = new RemoteExecutor();
            executor.Filter = ExtractFilter();

            return m_source.GetRemote(executor).GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        public Projector<TData, TTarget> Select<TTarget>(Expression<Func<TData, TTarget>> projection)
        {
            return new Projector<TData, TTarget>(this, projection);
        }
    }
}
